from .provider import Providers, Provider
from .kmindex import KmindexServerProvider

from kmviz.core.metadata import make_metadb_from_dict

PROVIDERS = {
    "kmindex-server": KmindexServerProvider
}

def make_provider_from_dict(name, d):
    if d["type"] not in PROVIDERS:
        raise RuntimeError("")

    p = PROVIDERS[d["type"]](name=name, **d["params"])
    metadb = make_metadb_from_dict(d["metadata"])
    p.attach_metadata(metadb)

    presets = { "map": {}, "plot": {}}

    if "presets" in d:
        pconf = d["presets"]
        for key in ["map", "plot"]:
            if key in pconf:
                for e in pconf[key]:
                    if not "priority" in e:
                        e["priority"] = False
                    presets[key][e["name"]] = e

        p.set_presets(presets)
    return p



import base64

from dash import Patch, no_update
import dash_mantine_components as dmc
from dash_extensions.enrich import Input, Output, State, html, dcc, Serverside, callback

from kmviz.ui.id_factory import kmviz_factory as kf
from kmviz.ui.patch import style_hide_patch, style_inline_patch
from kmviz.ui.components.store import ksfr
from kmviz.ui.utils import prevent_update_on_none

from kmviz.core.query import Query
from kmviz.core.io import parse_fastx, KmVizIOError
from kmviz.ui import state
from kmviz.ui.utils import make_select_data
from kmviz.ui.components.store import ksf
from kmviz.ui.components.select import kgsf
from dataclasses import dataclass

from io import StringIO
import pandas as pd

kif = kf.child("input")
kidf = kif.child("div")

def make_input_session():
    return [
        dmc.TextInput(
            id=kif("session"),
            required=True,
            label="Session ID",
            placeholder="kmviz-324bfabc-50e7-44ac-89d9-6460803d2f5e"
        ),
        dmc.Space(h=5),
        html.Div(id=kif("session-error")),
        dmc.Button("Load", id=kif("session-load"), style={"display":"inline"})
    ]

def make_input_session_callbacks():
    @callback(
        Input(kif("session-load"), "n_clicks"),
        State(kif("session"), "value"),
        Output(ksf("query-results"), "data"),
        Output(kgsf("provider"), "data"),
        Output(kgsf("provider"), "value"),
        Output(kgsf("query"), "data"),
        Output(kgsf("query"), "value"),
        Output(kf.sid("sidebar-layout"), "style"),
        Output(kif("session"), "error"),
        Output(kf.sid("session-id"), "data"),
        prevent_initial_callbacks=True,
        prevent_initial_call=True,
    )
    def load_session(n_clicks, session_id):
        if n_clicks:
            try:
                res = state.kmstate.get_result(session_id)
                return Serverside(res[0]), res[1], res[2], res[3], res[4], {"display":"none"}, no_update, session_id
            except:
                msg = "Session not found. Invalid session id, query still running, or results erased."
                return no_update, no_update, no_update, no_update, no_update, no_update, msg, None

def make_input_dataframe():
    upload_style = {
        'width': '80%%',
        'height': '40px',
        'lineHeight': '40px',
        'borderWidth': '1px',
        'borderStyle': 'dashed',
        'borderRadius': '5px',
        'textAlign': 'center',
        'margin': '10px',
        'align': 'center',
        'display': 'inherit'
    }

    return [
        dcc.Upload(
            id=kif("dataframe"),
            children=["Drop or ", html.A("Select a file")],
            style = upload_style
        ),
        html.Div(id=kif("dataframe-error"))
    ]

@dataclass
class _df_wrap:
    df: pd.DataFrame

def make_input_dataframe_callbacks():

    @callback(
        Input(kif("dataframe"), "filename"),
        Input(kif("dataframe"), "contents"),
        Output(ksf("query-results"), "data"),
        Output(kif("dataframe-error"), "children"),
        Output(kf.sid("sidebar-layout"), "style"),
        Output(kf.sid("index-panel"), "style"),
        Output(kf.sid("seq-panel"), "style"),
        Output(kgsf("provider"), "value"),
        Output(kgsf("query"), "value"),
        Output(kgsf("provider"), "style"),
        Output(kgsf("query"), "style"),
        Output("tab-select", "value"),


        prevent_initial_callbacks=True,
        prevent_initial_call=True,
    )
    def load_input_from_file(filename, contents):

        hide = {"display": "none"}

        prevent_update_on_none(filename, contents)

        content_type, data = contents.split(",")
        content = base64.b64decode(data).decode()

        try:
            df = None
            pdio = StringIO(content)
            if filename.endswith("csv") or filename.endswith("tsv"):
                try:
                    df = pd.read_csv(pdio, sep="\t")
                except:
                    raise KmVizIOError(f"Error while loading '{filename}'")

            key = "__kmviz_df_nogeo"
            if "Latitude" in df and "Longitude" in df:
                key = "__kmviz_df"

            res = {}
            res[key] = {}
            res[key][key] = _df_wrap(df)

            return Serverside(res), no_update, hide, hide, hide, key, key, hide, hide, "table"

        except KmVizIOError as e:
            message = dmc.Text(str(e), color="red", weight=500)
            return Serverside([]), message, no_update, no_update, no_update, no_update, no_update, no_update, no_update, no_update


def make_input_text():
    return [
        dmc.Textarea(
            id=kif("text"),
            label="Query sequence(s)",
            description="Fasta/Fastq format",
            placeholder=">Query\nACCGTAGCCTTAGAATTA",
            spellCheck=False,
            autosize=True,
            maxRows=4,
            required=True,
            style={"display":"inline"}
        ),
        dmc.Space(h=5),
        html.Div(id=kif("text-msg")),
        dmc.Button("Load", id=kif("text-load"), style={"display":"inline"})
    ]

def make_input_text_callbacks():

    @callback(
        Input(kif("text-load"), "n_clicks"),
        State(kif("text"), "value"),
        Output(ksfr("query-sequences"), "data"),
        Output(kif("text-load"), "style"),
        Output(kif("text"), "error"),
        Output(kif("text-msg"), "children"),
        Output(kif("select"), "style"),
        Output(kif("text"), "style"),
        prevent_initial_callbacks=True,
        prevent_initial_call=True,
    )
    def load_input_from_text(n_clicks, content):
        if n_clicks:
            try:
                queries = [Query(name, seq) for name, seq, _ in parse_fastx(content)]
                query = "queries" if len(queries) > 1 else "query"
                message = dmc.Text(f"🗎 raw ({len(queries)} {query})", weight=500)
                return Serverside(queries), style_hide_patch(), None, message, style_hide_patch(), style_hide_patch()
            except KmVizIOError as e:
                return Serverside([]), Patch(), str(e), None, no_update, style_inline_patch()

    @callback(
        Input(kif("text"), "value"),
        Output(kif("text-load"), "disabled"),
        prevent_initial_callbacks=True,
        prevent_initial_call=True,
    )
    def enable_load_button(value):
        if not value:
            return True
        return False

KMVIZ_UPLOAD_STYLE = {
    'width': '80%%',
    'height': '40px',
    'lineHeight': '40px',
    'borderWidth': '1px',
    'borderStyle': 'dashed',
    'borderRadius': '5px',
    'textAlign': 'center',
    'margin': '10px',
    'align': 'center',
    'display': 'inherit'
}


def make_input_file():

    return [
        dcc.Upload(
            id=kif("file"),
            children=["Drop or ", html.A("Select a file")],
            style = KMVIZ_UPLOAD_STYLE
        ),
        html.Div(id=kif("file-error"))
    ]

def make_input_file_callbacks():

    @callback(
        Input(kif("file"), "filename"),
        Input(kif("file"), "contents"),
        Output(ksfr("query-sequences"), "data"),
        Output(kif("file-error"), "children"),
        Output(kif("file"), "style"),
        Output(kif("file"), "filename"),
        Output(kif("select"), "style"),
        prevent_initial_callbacks=True,
        prevent_initial_call=True,
    )
    def load_input_from_file(filename, contents):
        prevent_update_on_none(filename, contents)

        content_type, data = contents.split(",")
        content = base64.b64decode(data).decode("utf-8")

        try:
            queries = [Query(name, seq) for name, seq, _ in parse_fastx(content)]
            query = "queries" if len(queries) > 1 else "query"
            message = dmc.Text(f"🗎 {filename} ({len(queries)} {query})", weight=500)
            return (
                Serverside(queries),
                message,
                style_hide_patch(),
                None,
                style_hide_patch()
            )
        except KmVizIOError as e:
            message = dmc.Text(f"🗎 {filename} -> Invalid format", color="red", weight=500)
            return Serverside([]), message, Patch(), None, style_inline_patch()

def make_input():
    hidden = { "display": "none" }
    show = { "display": "inline" }

    res = html.Div([
        dmc.Divider(size="sm", color="gray", label="INPUT", labelPosition="center"),
        dmc.Space(h=5),

        dmc.SegmentedControl(
            id=kif("select"),
            orientation="horizontal",
            size="sm",
            fullWidth=True,
            style={},
            color="#1C7ED6",
            data=make_select_data(["text", "file", "session"], True)
        ),


        dmc.Space(h=10),

        html.Div(make_input_text(), id=kidf("text"), style=show),
        html.Div(make_input_file(), id=kidf("file"), style=hidden),
        html.Div(make_input_session(), id=kidf("session"), style=hidden),
        #html.Div(make_input_dataframe(), id=kidf("df"), style=hidden),
    ])

    return res

def make_input_callbacks():

    @callback(
        Input(kif("select"), "value"),
        State(kidf.all, "id"),
        Output(kidf.all, "style"),
        prevent_initial_callbacks=True,
        prevent_initial_call=True,
    )
    def show_input(input_type, idx):
        prevent_update_on_none(input_type)
        styles = [{"display": "none"} for _ in range(len(idx))]

        if not input_type:
            return styles

        for index, id in enumerate(idx):
            if id["index"] == input_type:
                styles[index]["display"] = "inline"

        return styles

    make_input_text_callbacks()
    make_input_file_callbacks()
    make_input_session_callbacks()





from dash_extensions.enrich import (
    html,
    dcc,
)

import dash
from dash_iconify import DashIconify
import dash_mantine_components as dmc

from kmviz.ui import state

def make_instance():

    plugin = state.kmstate.instance_plugin()

    if plugin:
        return plugin.instance()
    #layout = html.Div([
    #    dcc.Markdown("""
    #    ## Welcome to kmviz ORA instance.

    #    Here is an example of what happens when kmviz runs with 'instance' plugins. In addition to the usual plugin functions, an instance plugin provides a homepage that may describe the instance content and provide some help. Here we load the 'kmviz_ora' plugin.
    #    (Note that this is not a proposal for the new ORA design, just me discovering modern css 🙂)
    #    """),
    #    html.Div([
    #    html.Div([
    #        html.Img(src="assets/water.webp", className="water"),
    #        html.Div([
    #            html.Img(src="assets/ship.png")
    #        ], className="ship"),
    #    ], className="ocean")
    #], className="hero"),

    #html.A(dmc.Button("Go to dashboard", leftIcon=DashIconify(icon="noto:rocket", width=20), style={"position":"fixed", "top": "50%", "left":"50%"}), href="/dashboard"),
    #])


    #return layout

plugin = state.kmstate.instance_plugin()

if plugin:
    dash.register_page(__name__, path="/", name=plugin.name(), title=plugin.name())

layout = make_instance()



from typing import Union, List, Any

from dash import Patch

def make_patch(*paths):
    p = Patch()
    for path in paths:
        p = p[path]
    return p

def patch_paths_from_id(id: Union[dict, str]) -> List[str]:
    if isinstance(id, str):
        paths = id.split("-")[-1].split("_")
    else:
        paths = id["index"].split("_")

    if len(paths) > 1:
        return paths[:-1], paths[-1]

    return None, paths[0]

def patch_value(patch: Patch, key: str, value: Any):
    patch[key] = value
    return patch

def patch_value_id(patch: Patch, id: Union[dict, str], value: Any):
    paths, key = patch_paths_from_id(id)

    if paths:
        for p in paths:
            patch = patch[p]

    return patch_value(patch, key, value)

def style_hide_patch():
    patch = Patch()
    patch["display"] = "none"
    return patch

def style_inline_patch():
    patch = Patch()
    patch["display"] = "inline"
    return patch



import pandas as pd
from typing import List, Set
from .db import MetaDB
import sqlite3

class SQLiteMetaDB(MetaDB):
    def __init__(self, path: str, idx: str, table: str, geodata: dict={}, **kwargs) -> None:
        super().__init__(idx, geodata)
        self.db = None
        self.path = path
        self.table = table

    def connect(self) -> None:
        self.db = sqlite3.connect(self.path)

    def query(self, keys: Set[str]) -> pd.DataFrame:
        keys_str = (f"'{x}'" for x in keys)
        return pd.read_sql(f"SELECT * FROM {self.table} WHERE {self.idx} IN ({','.join(keys_str)})", self.db)

    def df(self) -> pd.DataFrame:
        return pd.read_sql(f"SELECT * from '{self.table}'", self.db)

    def keys(self) -> List[str]:
        cur = self.db.execute(f"SELECT * from '{self.table}'")
        return list(map(lambda x: x[0], cur.description))




from dash_extensions.enrich import FileSystemBackend
from kmviz.core.provider import Providers, make_provider_from_dict
from kmviz.core.cache import result_cache_manager, callback_manager
from kmviz.core import KmVizError
from kmviz.core.plugin import installed_plugins, search_for_plugins
from kmviz.core.log import kmv_info
from kmviz.core.provider import PROVIDERS
from kmviz.core.metadata import METADBS
from importlib_resources import files
import shutil

class kState:
    def __init__(self):
        self._providers: Providers = Providers()
        self.dashboard_path = "/"
        self._manager = None
        self._cache = None
        self._plugins = {}
        self._external_css = []
        self._external_js = []
        self.plot_only = False
        self.backend = None

    def store_result(self, uid: str, results: tuple):
        self._cache[uid] = results
        self._cache.close()

    def get_result(self, uid: str) -> tuple:
        res = self._cache[uid]
        self._cache.close()
        return res

    def configure(self, config: dict):
        self._configure_plugins(config)
        self._configure_providers(config)
        self._configure_caches(config)

    @property
    def plugins(self):
        return self._plugins

    @property
    def providers(self):
        return self._providers

    @property
    def manager(self):
        return self._manager

    @property
    def css(self):
        return self._external_css

    @property
    def js(self):
        return self._external_js

    def instance_plugin(self):
        for name, p in self._plugins.items():
            if p.is_instance_plugin():
                return p
        return None

    def _configure_caches(self, config: dict):
        if "cache" not in config:
            raise KmVizError("'cache' section is missing in the configuration file.")

        if "manager" not in config["cache"]:
            raise KmVizError("'cache.manager' section is missing in the configuration file.")
        self._manager = callback_manager(config["cache"]["manager"])

        if "result" not in config["cache"]:
            raise KmVizError("'cache.result' section is missing in the configuration file.")

        if "backend" not in config["cache"]:
            self.backend = FileSystemBackend("file_system_backend")
        else:
            self.backend = FileSystemBackend(config["cache"]["backend"])

        self._init_cache(config["cache"]["result"])

    def _init_cache(self, params: dict):
        self._cache = result_cache_manager(params["params"])

    def _configure_plugins(self, config: dict):
        installed = installed_plugins()
        kmv_info(f"Installed plugins: {','.join(installed)}")

        if not "plugins" in config:
            self.dashboard_path = "/"
        else:
            self.dashboard_path = "/dashboard"
            self._plugins = search_for_plugins(config["plugins"])
            for name, plugin in self._plugins.items():
                self._external_css.extend(plugin.external_styles())
                self._external_js.extend(plugin.external_scripts())
                self._copy_plugin_assets(name)

                p_providers = plugin.providers()
                p_metadb = plugin.databases()

                if p_providers:
                    for pname, prov in p_providers:
                        if pname not in PROVIDERS:
                            kmv_info(f"Load external Provider '{pname}' from '{name}' plugin")
                            PROVIDERS[pname] = prov

                if p_metadb:
                    for pname, metadb in p_metadb:
                        if pname not in METADBS:
                            kmv_info(f"Load external MetaDB '{pname}' from '{name}' plugin")
                            METADBS[pname] = metadb

    def _copy_plugin_assets(self, plugin_name):
        p = files(plugin_name).joinpath("assets")
        main = str(files("kmviz").joinpath("assets"))
        if p.exists() and p.is_dir():
            shutil.copytree(p, f"{main}/_{plugin_name}_assets", dirs_exist_ok=True)

    def _configure_providers(self, config: dict):
        if "databases" not in config:
            raise KmVizError("'databases' section is missing in the configuration file.")

        for name, params in config["databases"].items():
            kmv_info(f"Load provider '{name}'")
            p = make_provider_from_dict(name, params)
            try:
                kmv_info(f"Init provider '{name}'")
                p.connect()
            except:
                raise KmVizError("Init fails for '{name}'")
            self._providers.add(p)

def init_state():
    global kmstate
    kmstate = kState()




from dash_extensions.enrich import DashLogger, Input, Output, State, html, dcc
from dash_extensions.enrich import Serverside, callback

from dash import no_update, Patch
from dash_iconify import DashIconify
import dash_mantine_components as dmc
import time
from kmviz.ui import state
from kmviz.ui.patch import style_hide_patch, style_inline_patch
from kmviz.ui.id_factory import kmviz_factory as kf
from kmviz.ui.utils import prevent_update_on_none, make_select_data
from kmviz.ui.components.store import ksf, ksfr
from kmviz.ui.components.select import kgsf
from kmviz.core.log import kmv_info
import time

import uuid

ksub = kf.child("submit")

def make_submit():
    res = html.Div([
        dmc.Center([
            dmc.Group([
                dmc.Button(
                    "Submit",
                    id=ksub.sid("button"),
                    leftIcon=DashIconify(icon="mdi:dna", width=20),
                    disabled=True
                ),
                html.A(
                    dmc.Button(
                        "Reset",
                        id=ksub.sid("refresh-button"),
                        leftIcon=DashIconify(icon="system-uicons:refresh", width=20),
                    ),
                    href=state.kmstate.dashboard_path
                ),
            ])
        ]),
        dmc.Space(h=10),
        html.Div([
            dmc.Progress(id=ksub.sid("progress"), value=0, size="xl", striped=True, animate=True)
        ], id=ksub.sid("progress-div"), style={"display": "none"}),
        dmc.NotificationsProvider([
            html.Div(id=kf.sid("notification")),
        ], limit=1),
    ])

    return res

def make_session_notification(action, uuid_str):
    msg="Here is your session id, use it to access your results later"
    notif = dmc.Notification(
        id="submit-notif",
        title=uuid_str,
        loading=False,
        message=msg,
        action=action,
        autoClose=False,
    )
    return notif

def make_submit_callbacks():

    @callback(
        Input(ksf("ready-state"), "data"),
        State(ksub.sid("button"), "n_clicks"),
        Output(ksub.sid("button"), "disabled")
    )
    def ready_to_submit(state, n_clicks):
        if n_clicks and n_clicks > 0:
            return True
        if all(state.values()):
            return False
        return True

    @callback(
        Output(ksf("query-results"), "data"),
        Output(kgsf("provider"), "data"),
        Output(kgsf("provider"), "value"),
        Output(kgsf("query"), "data"),
        Output(kgsf("query"), "value"),
        Output(kf.sid("session-id"), "data"),
        inputs=[
            Input(ksub.sid("button"), "n_clicks"),
            State(ksfr("query-sequences"), "data"),
            State(ksfr("provider-options"), "data"),
            State(ksfr("provider-active"), "data"),
        ],
        background=True,
        running=[
            (Output(ksub.sid("button"), "disabled"), True, True),
            (Output(ksub.sid("progress"), "animate"), True, False),
            (Output(ksub.sid("progress-div"), "style"), style_inline_patch(), style_inline_patch())
        ],
        progress=[Output(ksub.sid("progress"), "value"),
                  Output(ksub.sid("progress"), "label"),
                  Output(kf.sid("notification"), "children")],
        prevent_initial_call=True,
    )
    def submit(set_progress, n_clicks, queries, options, actives):
        prevent_update_on_none(n_clicks)

        nb_queries = len(queries)
        progress_pattern = "{n} / {total}"

        uuid_str = f"kmviz-{str(uuid.uuid4())}"

        kmv_info(f"⌛ {uuid_str}")
        nshow = make_session_notification("show", uuid_str)

        set_progress((0, progress_pattern.format(n=0, total=nb_queries), nshow))

        query_results = {}

        for i, query in enumerate(queries):
            result = state.kmstate.providers.query(query, actives, options, uuid_str)

            keys = list(result.keys())
            for key in keys:
                if isinstance(result[key], str):
                    del r[k]
            query_results[query.name] = result

            set_progress(((i / nb_queries) * 100, progress_pattern.format(n=i, total=nb_queries), nshow))

        default_query = queries[0].name
        default_provider = actives[0]

        state.kmstate.store_result(
            uuid_str,
            (query_results,
            make_select_data(actives),
            default_provider,
            make_select_data([query.name for query in queries]),
            default_query)
        )

        set_progress((100, "Done", nshow))

        kmv_info(f"✅ {uuid_str}")
        return (
            Serverside(query_results),
            make_select_data(actives),
            default_provider,
            make_select_data([query.name for query in queries]),
            default_query,
            uuid_str
        )




from dash_extensions.enrich import html, dcc, Input, Output, State, callback
from kmviz.ui.id_factory import kmviz_factory

from kmviz.ui import state

ksf = kmviz_factory.child("store")
ksfr = ksf.child("required")

def make_stores():

    default_options = {}
    for provider_name in state.kmstate.providers.list():
        default_options[provider_name] = {}
        for opt_name, opt in state.kmstate.providers.get(provider_name).options.items():
            default_options[provider_name][opt_name] = opt.value if opt.value else opt.default

    return html.Div([
        dcc.Store(id=ksfr("query-sequences"), data=[]),
        dcc.Store(id=ksfr("provider-options"), data=default_options),
        dcc.Store(id=ksfr("provider-active"), data=[]),
        dcc.Store(id=ksf("ready-state"), data={}),
        dcc.Store(id=ksf("query-results"), data={})
    ])

def make_stores_callbacks():

    @callback(
        Output(ksf("ready-state"), "data"),
        Input(ksfr.all, "data"),
        State(ksfr.all, "id"),
        prevent_intial_callbacks=True
    )
    def update_ready_state(stores, idx):
        return {
            idx[i]["index"]: store is not None and len(store)
            for i, store in enumerate(stores)
        }



from abc import ABC, abstractmethod
from typing import Union, Dict, Any, List

import pandas as pd

from dataclasses import dataclass
from kmviz.core.query import QueryResponse, Query
from kmviz.core.metadata.db import MetaDB

from .options import ProviderOption

class Provider(ABC):
    def __init__(self, name: str):
        self._name = name
        self.index_infos: Dict[str, Any] = None
        self.db: MetaDB = None
        self.options = {}
        self._presets = {}

    @abstractmethod
    def connect(self):
        """
        Initialization, like database login, should take place here
        """
        ...

    @abstractmethod
    def index_list(self) -> List[str]:
        """
        :returns: The list of sub-index names
        """
        ...

    @abstractmethod
    def query(self, query: Query, options: dict, idx: str) -> QueryResponse:
        """
        :param query: The query
        :param options: The user-defined options 
        :param idx: A unique id representing the query
        :returns: T
        """
        ...

    @abstractmethod
    def samples(self, index=None) -> Union[Dict[str, list], List[str]]:
        """
        :param index: The sub-index name
        :returns: The list of sample indexed by the sub-index, or a dict of list with one key per sub-index
        """
        ...

    @abstractmethod
    def kmer_size(self) -> int:
        """
        :returns: The kmer size for kmer-based index, 0 otherwise
        """
        ...

    @property
    def presets(self) -> dict:
        return self._presets

    @property
    def ready(self) -> bool:
        return True

    def options(self) -> Dict[str, ProviderOption]:
        return self.options

    def set_presets(self, presets: dict):
        self._presets = presets

    def name(self) -> str:
        return self._name

    def nb_samples(self, index=None) -> int:
        return len(self.samples(index))

    def infos(self) -> Dict[str, Any]:
        return self.index_infos

    def infos_df(self) -> pd.DataFrame:
        return pd.DataFrame.from_dict(self.index_infos)

    def attach_metadata(self, metadata: MetaDB) -> None:
        self.db = metadata

    def filter_metadata(self, query_metadata: pd.DataFrame) -> pd.DataFrame:
        return pd.merge(self.db.df(), query_metadata, on=[])

@dataclass
class _Opt:
    value: Any

class Providers:
    def __init__(self):
        self._providers = {}
        self._active = set()

    def add(self, provider: Provider) -> None:
        self._providers[provider.name()] = provider

    def list(self) -> List[str]:
        return list(self._providers.keys())

    def all(self):
        return self._providers

    def get(self, name: str):
        return self._providers[name]

    def query(self, query: Query, actives: list, options: dict, idx: str) -> Dict[str, QueryResponse]:
        res = {}
        for p in actives:
            opt = { k: _Opt(v) for k, v in options[p].items()}
            res[self._providers[p].name()] = self._providers[p].query(query, opt, idx)
        return res





from kmviz.core.provider import Provider
from kmviz.core.metadata.db import MetaDB
from kmviz.core import KmVizError
from typing import List, Tuple, Any, Union
import importlib
import pkgutil

plugin_prefix = "kmviz_"

class KmVizPlugin:

    def providers(self) -> List[Tuple[str, Provider]]:
        """
        :returns: The providers implemented by the plugin, as list of tuples <name,'Provider'>
        """
        return []

    def databases(self) -> List[Tuple[str, MetaDB]]:
        """
        :returns: The metadbs implemented by the plugin, as list of tuples <name,'MetaDB'>
        """
        return []

    def layouts(self) -> List[Tuple[str, Any, str]]:
        """
        :returns: The layouts implemented by the plugin, as list of tuples <name, dash_component, icon_name>
        """
        return []

    def external_scripts(self) -> List[Union[dict, str]]:
        """
        :returns: A list of js scripts to load, see https://dash.plotly.com/external-resources
        """
        return []

    def external_styles(self) -> List[Union[dict, str]]:
        """
        :returns: A list of css stylesheets to load, see https://dash.plotly.com/external-resources
        """
        return []

    def help(self) -> Any:
        """
        :returns: A Dash Component which will be displayed in the help tab.
        """
        return None

    def is_instance_plugin(self) -> bool:
        """
        :returns: True if the plugin is an instance plugin, False otherwise
        """
        return False

    def instance(self) -> Any:
        """
        :returns: A Dash Component which will be used as a homepage
        """
        return None

    def name(self) -> str:
        """
        :returns: The plugin name
        """
        return None

def installed_plugins() -> List[str]:
    res = []
    for finder, name, ispkg in pkgutil.iter_modules():
        if name.startswith("kmviz_"):
            res.append(name)
    return res

def search_for_plugins(modules: List[str]):
    plugins = {}

    for name in modules:
        if name.startswith("kmviz_"):
            try:
                module = importlib.import_module(name)
                plugins[name] = module.kmviz_plugin

                if not isinstance(plugins[name], KmVizPlugin):
                    raise KmVizError(f"Found module {name}, but 'kmviz_plugin' is not an instance of KmVizPlugin")
            except ImportError as e:
                raise KmVizError(f"Module '{name}' not found, installed plugins are [{','.join(installed_plugins())}]")

    return plugins





import dash_ag_grid as dag
from typing import Union

def make_ag_grid(id: Union[str, dict],
                 grid_opt: dict={},
                 col_def: dict={},
                 csv_export: dict={},
                 **kwargs):

    default_params = {
        "columnSize": "autoSize",
        "className": "ag-theme-balham"
    }

    default_col = {
        "filter": True,
        "sortable": True,
        "floatingFilter": True
    }

    default_opt = {
        "suppressMenuHide": False,
        "animateRows": True,
        "pagination": True,
        "loadingOverlayComponent": "CustomLoadingOverlay",
        "loadingOverlayComponentParams": {
            "loadingMessage": "Waiting for results..."
        },
        "rowSelection": "single",
    }

    default_export = {
        "fileName": "data.tsv",
        "columnSeparator": "\t"
    }

    return dag.AgGrid(
        id=id,
        defaultColDef={**default_col, **col_def},
        dashGridOptions={**default_opt, **grid_opt},
        csvExportParams={**default_export, **csv_export},
        **{**default_params, **kwargs},
    )




from dash import DiskcacheManager, CeleryManager

def callback_manager(params: dict):
    if params["type"] == "disk":
        return disk_callback_manager(params["params"])
    else:
        return celery_callback_manager(params["params"])

def disk_callback_manager(params) -> DiskcacheManager:
    import diskcache
    cache = diskcache.Cache(**params)
    return DiskcacheManager(cache)

def celery_callback_manager(params) -> CeleryManager:
    from celery import Celery
    app = Celery("kmviz", **params)
    return CeleryManager(app)

def result_cache_manager(params: dict):
    import diskcache
    cache = diskcache.Cache(**params)
    return cache



from dash_extensions.enrich import html, callback, Input, Output, dcc, State
import dash_mantine_components as dmc
from dash_iconify import DashIconify
from kmviz.ui import state
import pandas as pd

import dash_ag_grid as dag

from kmviz.ui.components.grid import make_ag_grid
from kmviz.ui.utils import prevent_update_on_none
from kmviz.ui.id_factory import kmviz_factory as kf
from kmviz.ui.components.select import kgsf

kindex = kf.child("index")

def make_index_layout():
    res = html.Div([
        dmc.Center([
            html.H2("Index informations")
        ]),
        make_ag_grid(
            kindex.sid("infos-grid"),
            style = {"height": None},
            grid_opt = {"domLayout": "autoHeight"}
        ),
        dmc.Space(h=20),
        dmc.Button(
            "Export",
            id=kindex.sid("infos-button"),
            disabled=True,
            leftIcon=DashIconify(icon="ph:export", width=20)
        ),
        dmc.Center([
            html.H2("Index metadata")
        ]),

        make_ag_grid(kindex.sid("metadata-grid")),
        dmc.Space(h=20),
        dmc.Group([
            dmc.Button(
                "Export",
                id=kindex.sid("metadata-button"),
                disabled=True,
                leftIcon=DashIconify(icon="ph:export", width=20)
            ),
            dmc.Button(
                "Remove filters",
                id=kindex.sid("metadata-rmf"),
                disabled=True,
                leftIcon=DashIconify(icon="ph:trash", width=20)
            ),
        ])
    ])

    return res

def make_index_layout_callbacks():

    @callback(
        Input(kindex.sid("metadata-rmf"), "n_clicks"),
        Output(kindex.sid("metadata-grid"), "filterModel"),
        prevent_initial_callbacks=True
    )
    def remove_metadata_table_filters(n_clicks):
        if n_clicks:
            return {}
        prevent_update_on_none(None)

    @callback(
        Input(kindex.sid("metadata-button"), "n_clicks"),
        Output(kindex.sid("metadata-grid"), "exportDataAsCsv")
    )
    def export_metadata(n_clicks):
        if n_clicks:
            return True
        prevent_update_on_none(None)

    @callback(
        Input(kindex.sid("infos-button"), "n_clicks"),
        Output(kindex.sid("infos-grid"), "exportDataAsCsv")
    )
    def export_metadata(n_clicks):
        if n_clicks:
            return True
        prevent_update_on_none(None)

    @callback(
        Input(kgsf("provider"), "value"),
        Output(kindex.sid("infos-button"), "disabled"),
        Output(kindex.sid("metadata-button"), "disabled"),
        Output(kindex.sid("metadata-rmf"), "disabled"),
        prevent_initial_callbacks=True
    )
    def enable_index_table_buttons(provider):
        prevent_update_on_none(provider)
        return False, False, False

    @callback(
        Input(kgsf("provider"), "value"),
        Output(kindex.sid("infos-grid"), "rowData"),
        Output(kindex.sid("infos-grid"), "columnDefs"),
        Output(kindex.sid("metadata-grid"), "rowData"),
        Output(kindex.sid("metadata-grid"), "columnDefs"),
        prevent_initial_callbacks=True
    )
    def update_index_table(provider):
        if provider == "__kmviz_df":
            prevent_update_on_none(None)

        p = state.kmstate.providers.get(provider)

        infos_rd = p.infos_df.to_dict("records")
        infos_f = [{"field": f} for f in list(p.infos_df)]

        meta_rd = p.db.df().to_dict("records")
        meta_f = [{"field": f} for f in list(p.db.df())]

        return infos_rd, infos_f, meta_rd, meta_f




import dash_mantine_components as dmc
from dash_extensions.enrich import html, callback, dcc, Input, Output, State, Serverside

from kmviz import __version__ as kmviz_version

from kmviz.ui.id_factory import kmviz_factory as kf
from kmviz.ui.components.select import make_select_provider, make_select_provider_callbacks
from kmviz.ui.components.input import make_input, make_input_callbacks, KMVIZ_UPLOAD_STYLE
from kmviz.ui.components.option import make_config, make_config_callbacks
from kmviz.ui.components.submit import make_submit, make_submit_callbacks
from kmviz.core.io import KmVizIOError

from kmviz.ui.utils import prevent_update_on_none
import base64
import pandas as pd
from io import StringIO, BytesIO


from kmviz.ui import state
klf = kf.child("layout")
ksb = klf.child("ksb")

def make_sidebar_layout():

    if state.kmstate.plot_only:
        res = html.Div([
            dcc.Store(id=kf.sid("plot-only"), data={}),
            dmc.Center([
                html.H1(f"kmviz v{kmviz_version}"),
            ]),
            html.H4(f"Plot only mode"),
            dmc.Divider(size="sm", color="gray", label="INPUT (csv/tsv/xlsx)", labelPosition="center"),
            dcc.Upload(
                id=ksb.sid("dataframe"),
                children=["Drop or ", html.A("Select a file")],
                style = KMVIZ_UPLOAD_STYLE
            ),
            dmc.Divider(size="sm", color="gray", label="CONFIG", labelPosition="center"),
            dmc.Group([
                dmc.TextInput(
                    label="Index column",
                    id=ksb.sid("index-column"),
                    style = {"width": "100px"}
                ),
                dmc.TextInput(
                    label="Separator",
                    id=ksb.sid("separator"),
                    style = {"width": "100px"}
                )
            ]),
            dmc.Group([
                dmc.TextInput(
                    label="Latitude column",
                    id=ksb.sid("latitude"),
                    style = {"width": "100px"}
                ),
                dmc.TextInput(
                    label="Longitude column",
                    id=ksb.sid("longitude"),
                    style = {"width": "100px"}
                )
            ]),
            dmc.Space(h=5),
            html.Div(id=ksb.sid("dataframe-error")),
            dmc.Space(h=5),
            dmc.Button(
                "Load",
                id=ksb.sid("df-button"),
            ),

        ], className="sidebar", id=ksb.sid("div"))
    else:
        res = html.Div([
            dcc.Store(id=kf.sid("plot-only"), data={}),
            dmc.Center([
                html.H1(f"kmviz v{kmviz_version}"),
            ]),
            dmc.Space(h=5),
            make_select_provider(),
            dmc.Space(h=20),
            make_input(),
            dmc.Space(h=20),
            make_config(),
            dmc.Space(h=20),
            make_submit()
        ], className="sidebar", id=ksb.sid("div"))

    return res

def make_sidebar_layout_callbacks():

    if state.kmstate.plot_only:
        @callback(
            Input(ksb.sid("dataframe"), "filename"),
            Output(ksb.sid("dataframe-error"), "children"),
        )
        def show_filename(filename):
            return dmc.Text(f"🗎 {filename}", weight=500)

        @callback(
            Input(ksb.sid("df-button"), "n_clicks"),
            State(ksb.sid("dataframe"), "filename"),
            State(ksb.sid("dataframe"), "contents"),
            State(ksb.sid("separator"), "value"),
            State(ksb.sid("index-column"), "value"),
            State(ksb.sid("latitude"), "value"),
            State(ksb.sid("longitude"), "value"),
            Output(kf.sid("plot-only"), "data"),
            Output(ksb.sid("dataframe-error"), "children"),

            prevent_initial_callbacks=True,
            prevent_initial_call=True,
        )
        def load_df_file(button, filename, contents, sep, index, lat, lon):

            hide = {"display": "none"}
            prevent_update_on_none(filename, contents)
            content_type, data = contents.split(",")
            content = base64.b64decode(data)

            try:
                df = None
                if filename.endswith("csv") or filename.endswith("tsv"):
                    pdio = StringIO(content.decode("utf-8"))
                    try:
                        if sep:
                            df = pd.read_csv(pdio, sep=sep)
                        else:
                            df = pd.read_csv(pdio)
                    except:
                        raise KmVizIOError(f"Error while loading '{filename}'")
                else:
                    pdio = BytesIO(content)
                    try:
                        df = pd.read_excel(pdio)
                    except:
                        raise KmVizIOError(f"Error while loading '{filename}'")

                res = {}

                if not index or index not in list(df):
                    raise KmVizIOError(f"'{index}' not found in dataframe.")

                df.rename(columns={index : "ID"}, inplace=True)
                res["df"] = df

                res["geodata"] = None
                if lat and lon:
                    if lat in list(df) and lon in list(df):
                        res["geodata"] = { "latitude": lat, "longitude": lon}
                    else:
                        raise KmVizIOError(f"'{lat}' or '{lon}' not found in dataframe")

                return Serverside(res), dmc.Text(f"🗎 {filename}", weight=500)
            except KmVizIOError as e:
                return Serverside(res), dmc.Text(str(e), color="red", weight=500)

    else:
        make_select_provider_callbacks()
        make_input_callbacks()
        make_config_callbacks()
        make_submit_callbacks()
